"""A set of helper utilities."""

import errno
import os
import logging

LOGGER = logging.getLogger(__name__)


def make_folder(folder):
    """Helper method to sanely create folders."""
    if not os.path.exists(folder):
        # Still try and except, since their may be race conditions.
        try:
            os.makedirs(folder)
        except OSError as error:
            if error.errno != errno.EEXIST:
                raise



"""default driver for the network_importer."""
# pylint: disable=raise-missing-from,too-many-arguments

import os
import socket
import jinja2

from netmiko.ssh_exception import NetmikoAuthenticationException, NetmikoTimeoutException
from nornir.core.exceptions import NornirSubTaskError
from nornir.core.task import Result, Task
from nornir_jinja2.plugins.tasks import template_file
from nornir_napalm.plugins.tasks import napalm_get
from nornir_netmiko.tasks import netmiko_send_command
from netutils.config.compliance import compliance
from netutils.config.clean import clean_config, sanitize_config
from netutils.ip import is_ip
from netutils.dns import is_fqdn_resolvable
from netutils.ping import tcp_ping

from nornir_nautobot.exceptions import NornirNautobotException
from nornir_nautobot.utils.helpers import make_folder


RUN_COMMAND_MAPPING = {
    "default": "show run",
    "cisco_nxos": "show run",
    "cisco_ios": "show run",
    "cisco_xr": "show run",
    "juniper_junos": "show configuration | display set",
    "arista_eos": "show run",
}


class NautobotNornirDriver:
    """Default collection of Nornir Tasks based on Napalm."""

    @staticmethod
    def get_config(task: Task, logger, obj, backup_file: str, remove_lines: list, substitute_lines: list) -> Result:
        """Get the latest configuration from the device.

        Args:
            task (Task): Nornir Task.
            logger (NornirLogger): Custom NornirLogger object to reflect job_results (via Nautobot Jobs) and Python logger.
            obj (Device): A Nautobot Device Django ORM object instance.
            backup_file (str): The file location of where the back configuration should be saved.
            remove_lines (list): A list of regex lines to remove configurations.
            substitute_lines (list): A list of dictionaries with to remove and replace lines.

        Returns:
            Result: Nornir Result object with a dict as a result containing the running configuration
                { "config: <running configuration> }
        """
        logger.log_debug(f"Executing get_config for {task.host.name} on {task.host.platform}")

        # TODO: Find standard napalm exceptions and account for them
        try:
            result = task.run(task=napalm_get, getters=["config"], retrieve="running")
        except NornirSubTaskError as exc:
            logger.log_failure(obj, f"Failed with a unknown issue. `{exc.result.exception}`")
            raise NornirNautobotException()

        if result[0].failed:
            logger.log_failure(obj, f"Failed with a unknown issue. `{str(result.exception)}`")
            return result

        running_config = result[0].result.get("config", {}).get("running", None)
        if remove_lines:
            logger.log_debug("Removing lines from configuration based on `remove_lines` definition")
            running_config = clean_config(running_config, remove_lines)

        if substitute_lines:
            logger.log_debug("Substitute lines from configuration based on `substitute_lines` definition")
            running_config = sanitize_config(running_config, substitute_lines)

        make_folder(os.path.dirname(backup_file))

        with open(backup_file, "w") as filehandler:
            filehandler.write(running_config)
        return Result(host=task.host, result={"config": running_config})

    @staticmethod
    def check_connectivity(task: Task, logger, obj) -> Result:
        """Check the connectivity to a network device.

        Args:
            task (Task): Nornir Task.
            logger (NornirLogger): Custom NornirLogger object to reflect job_results (via Nautobot Jobs) and Python logger.
            obj (Device): A Nautobot Device Django ORM object instance.

        Returns:
            Result: Nornir Result object.
        """
        if is_ip(task.host.hostname):
            ip_addr = task.host.hostname
        else:
            if not is_fqdn_resolvable(task.host.hostname):
                logger.log_failure(obj, "not an IP or resolvable.")
                raise NornirNautobotException("not an IP or resolvable.")
            ip_addr = socket.gethostbyname(task.host.hostname)

        # TODO: Allow port to be configurable
        port = 22
        if not tcp_ping(ip_addr, port):
            logger.log_failure(obj, f"Attempting to connect to IP: {ip_addr} and port: {port} failed.")
            raise NornirNautobotException(f"Attempting to connect to IP: {ip_addr} and port: {port} failed.")
        if not task.host.username:
            logger.log_failure(obj, "There was no username defined, preemptively failed.")
            raise NornirNautobotException("There was no username defined, preemptively failed.")
        if not task.host.password:
            logger.log_failure(obj, "There was no password defined, preemptively failed.")
            raise NornirNautobotException("There was no password defined, preemptively failed.")

        return Result(host=task.host)

    @staticmethod
    def compliance_config(
        task: Task, logger, obj, features: str, backup_file: str, intended_file: str, platform: str
    ) -> Result:
        """Compare two configurations against each other.

        Args:
            task (Task): Nornir Task.
            logger (NornirLogger): Custom NornirLogger object to reflect job_results (via Nautobot Jobs) and Python logger.
            obj (Device): A Nautobot Device Django ORM object instance.
            features (dict): A dictionary describing the configurations required.
            backup_file (str): The file location of where the back configuration should be saved.
            intended_file (str):  The file location of where the intended configuration should be saved.
            platform (str): The platform slug of the device.

        Returns:
            Result: Nornir Result object with a feature_data key of the compliance data.
        """
        if not os.path.exists(backup_file):
            logger.log_failure(obj, f"Backup file Not Found at location: `{backup_file}`")
            raise NornirNautobotException()

        if not os.path.exists(intended_file):
            logger.log_failure(obj, f"Intended config file NOT Found at location: `{intended_file}`")
            raise NornirNautobotException()

        try:
            feature_data = compliance(features, backup_file, intended_file, platform)
        except Exception as error:  # pylint: disable=broad-except
            logger.log_failure(obj, f"UNKNOWN Failure of: {str(error)}")
            raise NornirNautobotException()
        return Result(host=task.host, result={"feature_data": feature_data})

    @staticmethod
    def generate_config(
        task: Task, logger, obj, jinja_template: str, jinja_root_path: str, output_file_location: str
    ) -> Result:
        """A small wrapper around template_file Nornir task.

        Args:
            task (Task): Nornir Task.
            logger (NornirLogger): Custom NornirLogger object to reflect job_results (via Nautobot Jobs) and Python logger.
            obj (Device): A Nautobot Device Django ORM object instance.
            jinja_template (str): The file location of the actual Jinja template.
            jinja_root_path (str): The file folder where the file will be saved to.
            output_file_location (str): The filename where the file will be saved to.

        Returns:
            Result: Nornir Result object.
        """
        try:
            filled_template = task.run(
                **task.host.data,
                task=template_file,
                template=jinja_template,
                path=jinja_root_path,
            )[0].result
        except NornirSubTaskError as exc:
            if isinstance(exc.result.exception, jinja2.exceptions.UndefinedError):  # pylint: disable=no-else-raise
                logger.log_failure(
                    obj,
                    f"There was a jinja2.exceptions.UndefinedError error: ``{str(exc.result.exception)}``",
                )
                raise NornirNautobotException()
            elif isinstance(exc.result.exception, jinja2.TemplateSyntaxError):
                logger.log_failure(
                    obj,
                    f"There was a jinja2.TemplateSyntaxError error: ``{str(exc.result.exception)}``",
                )
                raise NornirNautobotException()
            elif isinstance(exc.result.exception, jinja2.TemplateNotFound):
                logger.log_failure(
                    obj,
                    f"There was an issue finding the template and a jinja2.TemplateNotFound error was raised: ``{str(exc.result.exception)}``",
                )
                raise NornirNautobotException()
            elif isinstance(exc.result.exception, jinja2.TemplateError):
                logger.log_failure(obj, f"There was an issue general Jinja error: ``{str(exc.result.exception)}``")
                raise NornirNautobotException()
            raise

        make_folder(os.path.dirname(output_file_location))
        with open(output_file_location, "w") as filehandler:
            filehandler.write(filled_template)
        return Result(host=task.host, result={"config": filled_template})


class NetmikoNautobotNornirDriver(NautobotNornirDriver):
    """Default collection of Nornir Tasks based on Netmiko."""

    @staticmethod
    def get_config(task: Task, logger, obj, backup_file: str, remove_lines: list, substitute_lines: list) -> Result:
        """Get the latest configuration from the device using Netmiko.

        Args:
            task (Task): Nornir Task.
            logger (NornirLogger): Custom NornirLogger object to reflect job_results (via Nautobot Jobs) and Python logger.
            obj (Device): A Nautobot Device Django ORM object instance.
            remove_lines (list): A list of regex lines to remove configurations.
            substitute_lines (list): A list of dictionaries with to remove and replace lines.

        Returns:
            Result: Nornir Result object with a dict as a result containing the running configuration
                { "config: <running configuration> }
        """
        logger.log_debug(f"Executing get_config for {task.host.name} on {task.host.platform}")
        command = RUN_COMMAND_MAPPING.get(task.host.platform, RUN_COMMAND_MAPPING["default"])

        try:
            result = task.run(task=netmiko_send_command, command_string=command)
        except NornirSubTaskError as exc:
            if isinstance(exc.result.exception, NetmikoAuthenticationException):
                logger.log_failure(obj, f"Failed with an authentication issue: `{exc.result.exception}`")
                raise NornirNautobotException()

            if isinstance(exc.result.exception, NetmikoTimeoutException):
                logger.log_failure(obj, f"Failed with a timeout issue. `{exc.result.exception}`")
                raise NornirNautobotException()

            logger.log_failure(obj, f"Failed with an unknown issue. `{exc.result.exception}`")
            raise NornirNautobotException()

        if result[0].failed:
            return result

        running_config = result[0].result

        # Primarily seen in Cisco devices.
        if "ERROR: % Invalid input detected at" in running_config:
            logger.log_failure(obj, "Discovered `ERROR: % Invalid input detected at` in the output")
            raise NornirNautobotException()

        if remove_lines:
            logger.log_debug("Removing lines from configuration based on `remove_lines` definition")
            running_config = clean_config(running_config, remove_lines)
        if substitute_lines:
            logger.log_debug("Substitute lines from configuration based on `substitute_lines` definition")
            running_config = sanitize_config(running_config, substitute_lines)

        make_folder(os.path.dirname(backup_file))

        with open(backup_file, "w") as filehandler:
            filehandler.write(running_config)
        return Result(host=task.host, result={"config": running_config})



"""Used to intialize the dispatcher."""
# pylint: disable=raise-missing-from

import importlib
import logging

from nornir.core.task import Result, Task
from nornir_nautobot.exceptions import NornirNautobotException

LOGGER = logging.getLogger(__name__)

_DEFAULT_DRIVERS_MAPPING = {
    "default": "nornir_nautobot.plugins.tasks.dispatcher.default.NautobotNornirDriver",
    "default_netmiko": "nornir_nautobot.plugins.tasks.dispatcher.default.NetmikoNautobotNornirDriver",
    "cisco_nxos": "nornir_nautobot.plugins.tasks.dispatcher.cisco_nxos.NautobotNornirDriver",
    "cisco_ios": "nornir_nautobot.plugins.tasks.dispatcher.cisco_ios.NautobotNornirDriver",
    "cisco_xr": "nornir_nautobot.plugins.tasks.dispatcher.cisco_xr.NautobotNornirDriver",
    "juniper_junos": "nornir_nautobot.plugins.tasks.dispatcher.juniper_junos.NautobotNornirDriver",
    "arista_eos": "nornir_nautobot.plugins.tasks.dispatcher.arista_eos.NautobotNornirDriver",
}


def dispatcher(task: Task, method: str, logger, obj, *args, **kwargs) -> Result:
    """Helper Task to retrieve a given Nornir task for a given platform.

    Args:
        task (Nornir Task):  Nornir Task object.
        method (str):  The string value of the method to dynamically find.

    Returns:
        Result: Nornir Task result.
    """
    if kwargs.get("default_drivers_mapping"):
        default_drivers_mapping = kwargs["default_drivers_mapping"]
        del kwargs["default_drivers_mapping"]
    else:
        default_drivers_mapping = _DEFAULT_DRIVERS_MAPPING

    logger.log_debug(f"Executing dispatcher for {task.host.name} ({task.host.platform})")

    # Get the platform specific driver, if not available, get the default driver
    driver = default_drivers_mapping.get(task.host.platform, default_drivers_mapping.get("default"))
    logger.log_debug(f"Found driver {driver}")

    if not driver:
        logger.log_failure(obj, f"Unable to find the driver for {method} for platform: {task.host.platform}")
        raise NornirNautobotException()

    module_name, class_name = driver.rsplit(".", 1)
    driver_class = getattr(importlib.import_module(module_name), class_name)

    if not driver_class:
        logger.log_failure(obj, f"Unable to locate the class {driver}")
        raise NornirNautobotException()

    try:
        driver_task = getattr(driver_class, method)
    except AttributeError:
        logger.log_failure(obj, f"Unable to locate the method {method} for {driver}")
        raise NornirNautobotException()

    result = task.run(task=driver_task, logger=logger, obj=obj, *args, **kwargs)

    return Result(
        host=task.host,
        result=result,
    )



"""network_importer driver for cisco IOS-XR."""

from .default import NetmikoNautobotNornirDriver as DefaultNautobotNornirDriver


class NautobotNornirDriver(DefaultNautobotNornirDriver):
    """Driver for Cisco IOS-XR."""



"""BaseProcessor for the nornir."""
import logging

from nornir.core.inventory import Host
from nornir.core.task import AggregatedResult, MultiResult, Task

LOGGER = logging.getLogger(__name__)


class BaseProcessor:
    """Base Processor for nornir."""

    task_name = "'no task defined'"

    def task_started(self, task: Task) -> None:
        """Boilerplate Nornir processor for task_started."""

    def task_completed(self, task: Task, result: AggregatedResult) -> None:
        """Boilerplate Nornir processor for task_completed."""

    def task_instance_started(self, task: Task, host: Host) -> None:
        """Boilerplate Nornir processor for task_instance_started."""

    def task_instance_completed(self, task: Task, host: Host, result: MultiResult) -> None:
        """Boilerplate Nornir processor for task_instance_completed."""

    def subtask_instance_started(self, task: Task, host: Host) -> None:
        """Boilerplate Nornir processor for subtask_instance_started."""

    def subtask_instance_completed(self, task: Task, host: Host, result: MultiResult) -> None:
        """Boilerplate Nornir processor for subtask_instance_completed."""


class BaseLoggingProcessor(BaseProcessor):
    """Base Processor with logging for nornir."""

    def task_started(self, task: Task) -> None:
        """Boilerplate Nornir processor for task_started with logging."""
        LOGGER.info("%s | Task started", task.name)

    def task_completed(self, task: Task, result: AggregatedResult) -> None:
        """Boilerplate Nornir processor for task_completed with logging."""
        LOGGER.info("%s | Task task_completed", task.name)

    def task_instance_started(self, task: Task, host: Host) -> None:
        """Boilerplate Nornir processor for task_instance_started with logging."""
        LOGGER.info("%s | Task instance %s has started", host.name, task.name)

    def task_instance_completed(self, task: Task, host: Host, result: MultiResult) -> None:
        """Boilerplate Nornir processor for task_instance_completed with logging."""
        LOGGER.info("%s | Task instance %s has completed", host.name, task.name)

    def subtask_instance_started(self, task: Task, host: Host) -> None:
        """Boilerplate Nornir processor for subtask_instance_started with logging."""
        LOGGER.info("%s | Task instance subtask %s has started.", task.host.name, task.name)

    def subtask_instance_completed(self, task: Task, host: Host, result: MultiResult) -> None:
        """Boilerplate Nornir processor for subtask_instance_completed with logging."""
        LOGGER.info("%s | Task instance subtask %s has completed.", task.host.name, task.name)



"""GetConfig processor for the network_importer."""

import hashlib
import logging
import os

from pathlib import Path

from nornir.core.inventory import Host
from nornir.core.task import AggregatedResult, MultiResult, Task

from . import BaseProcessor

# TODO: Adjust to work with a configuration setting
CONFIG = None
LOGGER = logging.getLogger(__name__)


class GetConfig(BaseProcessor):
    """GetConfig processor for the network_importer."""

    task_name = "get_config"
    config_extension = "txt"

    def __init__(self) -> None:
        """Initialize the processor and ensure some variables are properly initialized."""
        self.current_md5 = dict()
        self.previous_md5 = dict()
        self.config_filename = dict()
        self.config_dir = None
        self.existing_config_hostnames = None

    def task_started(self, task: Task) -> None:
        """Execute some house keeping item at the beginning at the execution.

        Before Update all the config file:
        - ensure that the configs directory exist
        - check what config files are already present

        Args:
            task (Task): Nornir Task
        """
        if not os.path.isdir(CONFIG.main.configs_directory):
            os.mkdir(CONFIG.main.configs_directory)
            LOGGER.debug("Configs directory created at %s", CONFIG.main.configs_directory)

        self.config_dir = CONFIG.main.configs_directory + "/configs"

        if not os.path.isdir(self.config_dir):
            os.mkdir(self.config_dir)
            LOGGER.debug("Configs directory created at %s", self.config_dir)

        # Save the hostnames associated with all existing configurations before we start the update process
        self.existing_config_hostnames = [
            f.split(f".{self.config_extension}")[0] for f in os.listdir(self.config_dir) if f.endswith(".txt")
        ]

    def task_completed(self, task: Task, result: AggregatedResult) -> None:
        """At the end, remove all configs files that have not been updated."""
        if len(self.existing_config_hostnames) > 0:
            LOGGER.info("Will delete %s config(s) that have not been updated", len(self.existing_config_hostnames))

            for hostname in self.existing_config_hostnames:
                os.remove(os.path.join(self.config_dir, f"{hostname}.{self.config_extension}"))

    def subtask_instance_started(self, task: Task, host: Host) -> None:
        """Before getting the new configuration, check if a configuration already exist and calculate its md5.

        Args:
            task (Task): Nornir Task
            host (Host): Nornir Host
        """
        if task.name != self.task_name:
            return

        self.config_filename[host.name] = f"{self.config_dir}/{task.host.name}.{self.config_extension}"

        if os.path.exists(self.config_filename[host.name]):
            current_config = Path(self.config_filename[host.name]).read_text()
            self.previous_md5[host.name] = hashlib.md5(current_config.encode("utf-8")).hexdigest()  # nosec

    def subtask_instance_completed(self, task: Task, host: Host, result: MultiResult) -> None:
        """Verify the configuration returned and store it to disk.

        After collecting the configuration for each host.
        - Inspect the result and ensure the configuration is valid
        - If the configuration is valid, check the md5
        - if the MD5 is different, write the new configuration to disk

        Args:
            task (Task): Nornir Task
            host (Host): Nornir Host
            result (MultiResult): Nornir MultiResult
        """
        if task.name != self.task_name:
            return

        if result[0].failed:
            if result[0].exception:
                LOGGER.warning("%s | %s", task.host.name, result[0].exception)
            else:
                LOGGER.warning("%s | Something went wrong while trying to update the configuration ", task.host.name)
            host.data["status"] = "fail-other"
            return

        conf = result[0].result.get("config", None)

        if not conf:
            LOGGER.warning("%s | No configuration return ", task.host.name)
            host.data["status"] = "fail-other"
            return

        # Count the number of lines in the config file, if less than 10 report an error
        # mostlikely something went wrong while pulling the config
        if conf.count("\n") < 10:
            LOGGER.warning("%s | Less than 10 configuration lines returned", task.host.name)
            host.data["status"] = "fail-other"
            return

        if host.name in self.existing_config_hostnames:
            self.existing_config_hostnames.remove(host.name)

        # Save configuration to file and verify the new MD5
        with open(self.config_filename[host.name], "w") as config_:
            config_.write(conf)

        host.data["has_config"] = True

        self.current_md5[host.name] = hashlib.md5(conf.encode("utf-8")).hexdigest()  # nosec
        # changed = False

        if host.name in self.previous_md5 and self.previous_md5[host.name] == self.current_md5[host.name]:
            LOGGER.info("%s | Latest config file already present ...", task.host.name)
        else:
            LOGGER.info("%s | Configuration file updated", task.host.name)
            # changed = True



"""Utilities used to mock other objects."""

from urllib3 import connectionpool, poolmanager


def patch_http_connection_pool(**constructor_kwargs):
    """This allows to override the default parameters of the HTTPConnectionPool constructor.

    For example, to increase the poolsize to fix problems
    with "HttpConnectionPool is full, discarding connection"
    call this function with maxsize=16 (or whatever size
    you want to give to the connection pool)

    Args:
      **constructor_kwargs:
    """

    class MyHTTPConnectionPool(connectionpool.HTTPConnectionPool):
        """Class to increase the size of the HTTP Connection pool."""

        def __init__(self, *args, **kwargs):
            """Initialize the HTTP Connection pool."""
            kwargs.update(constructor_kwargs)
            super().__init__(*args, **kwargs)

    poolmanager.pool_classes_by_scheme["http"] = MyHTTPConnectionPool
